//---------------------------------------------------------------------------//
//  1100101 |_ mov #0, r4         __                                         //
//     11   |_ <0xb380 %5c4>     / _|_ _____ ___                             //
//     0110 |_ 3.50 -> 3.60     |  _\ \ / _ (_-<                             //
//          |_ base# + offset   |_| /_\_\___/__/                             //
//---------------------------------------------------------------------------//

#include <fxos/passes/cfg.h>
#include <fxos/disassembly.h>
#include <fxos/util/log.h>
#include <fxos/binary.h>
#include <cassert>

namespace FxOS {

CfgPass::CfgPass(Binary &binary):
    InstructionPass(binary), m_claimedInstructions {}, m_pcrel {binary}
{
    this->setAllowDiscovery(true);
}

bool CfgPass::analyzeInstruction(uint32_t pc, OldInstruction &i)
{
    /* Don't explore successors if the instruction cannot be decoded, not
       even pc+2. This will prevent wild overshoot. */
    if(!i.inst) {
        FxOS_log(ERR, "invalid instruction at 0x%08x: 0x%04x", pc, i.opcode);
        return false;
    }

    m_claimedInstructions.insert(pc);

    /* Compute the jump target for jump instructions. This is easy because
       they are all trivially computable. (...If they are not we dub them
       "terminal" to avoid the computation!) */
    uint32_t jmptarget = 0xffffffff;

    if(i.inst->isAnyJump()) {
        auto &args = i.inst->args;

        if(i.inst->arg_count != 1 || args[0].kind != AsmArgument::PcJump) {
            FxOS_log(ERR, "invalid jump instruction at 0x%08x", pc);
            return false;
        }

        jmptarget = (pc + 4) + args[0].disp;

        /* Make the target of the jump a leader */
        // TODO: Use Binary instructions
        OldInstruction &target
            = *m_binary.vspace().disasm.getInstructionAt(jmptarget, true);
        target.leader = true;

        /* Check that it's not in a delay slot */
        if(target.delayslot)
            throw std::logic_error(format(
                "0x%08x jumps into 0x%08x, which is "
                "a delay slot - this is unsupported by fxos and will produce "
                "garbage analysis! (x_x)",
                pc, jmptarget));
    }

    /* If this instruction is in a delay slot, check its type. A valid
       delay slot has no branching properties on its own, so nothing new to
       set in the properties.  */
    if(i.delayslot) {
        if(!i.inst->isValidDelaySlot()) {
            FxOS_log(ERR, "invalid delay slot at 0x%08x", pc);
            return false;
        }
    }
    /* If it has a delay slot, create it at the next instruction */
    else if(i.inst->hasDelaySlot()) {
        // TODO: Use Binary instructions
        OldInstruction &slot
            = *m_binary.vspace().disasm.getInstructionAt(pc + 2, true);
        if(slot.leader)
            throw std::logic_error(format(
                "0x%08x is a leader and also a delay"
                " slot - this is unsupported by fxos and will produce garbage "
                "analysis! (x_x)",
                pc + 2));
        if(!slot.inst->isValidDelaySlot()) {
            FxOS_log(ERR, "invalid delay slot at 0x%08x", pc + 2);
            return false;
        }

        slot.delayslot = true;
        slot.terminal = i.inst->isReturn();
        slot.jump = i.inst->isUnconditionalJump();
        slot.condjump = i.inst->isConditionalJump();
        slot.jmptarget = jmptarget;
    }
    /* Otherwise, use standard properties */
    else if(!i.inst->hasDelaySlot()) {
        i.terminal = i.inst->isReturn();
        i.jump = i.inst->isUnconditionalJump();
        i.condjump = i.inst->isConditionalJump();
        i.jmptarget = jmptarget;
    }

    m_pcrel.analyzeInstruction(pc, i);
    return true;
}

bool CfgPass::exploreFunction(uint32_t pc)
{
    m_lastFunction = pc;
    m_claimedInstructions.clear();

    // TODO: Use Binary functions
    OldFunction *func = m_binary.vspace().disasm.getOrCreateFunctionAt(pc);
    if(!this->analyzeFunction(pc))
        return false;

    RelConstDomain RCD;

    /* Look for call targets */
    for(uint32_t pc: m_claimedInstructions) {
        // TODO: Use Binary instructions
        OldInstruction const *ci
            = m_binary.vspace().disasm.getInstructionAt(pc);
        if(!ci)
            continue;
        AsmInstruction const &i = *ci->inst;

        /* Find function call instructions */
        if(i.isReturn() || !i.isCall() || i.arg_count < 1)
            continue;

        /* The target must be known */
        if(!RCD.is_constant(ci->args[0].location))
            continue;

        uint32_t target = RCD.constant_value(ci->args[0].location);
        auto &v = func->callTargets;

        if(std::find(v.begin(), v.end(), target) == v.end())
            func->callTargets.push_back(target);
    }

    return true;
}

} /* namespace FxOS */
